/*
 * synergy -- mouse and keyboard sharing utility
 * Copyright (C) 2002 Chris Schoeneman, Nick Bolton, Sorin Sbarnea
 * 
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * found in the file COPYING that should have accompanied this file.
 * 
 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#if defined(_MSC_VER)
#pragma warning(disable: 4786)		// identifier truncated
#pragma warning(disable: 4514)		// unreferenced inline
#pragma warning(disable: 4710)		// not inlined
#pragma warning(disable: 4663)		// C++ change, template specialization
#pragma warning(disable: 4503)		// decorated name length too long
#pragma warning(push, 3)
#pragma warning(disable: 4018)		// signed/unsigned mismatch
#pragma warning(disable: 4284)
#pragma warning(disable: 4146)		// unary minus on unsigned value
#pragma warning(disable: 4127)		// conditional expression is constant
#pragma warning(disable: 4701)		// variable possibly used uninitialized
#endif
