//Microsoft Developer Studio generated resource script.
//
#include "resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include <winresrc.h>
#if !defined(IDC_STATIC)
#define IDC_STATIC (-1)
#endif

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (U.S.) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
#ifdef _WIN32
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
#pragma code_page(1252)
#endif //_WIN32

#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE DISCARDABLE 
BEGIN
    "resource.h\0"
END

2 TEXTINCLUDE DISCARDABLE 
BEGIN
    "#include <winresrc.h>\r\n"
    "\0"
END

3 TEXTINCLUDE DISCARDABLE 
BEGIN
    "\r\n"
    "\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_MAIN DIALOG DISCARDABLE  32768, 0, 300, 199
STYLE DS_MODALFRAME | WS_MINIMIZEBOX | WS_CAPTION | WS_SYSMENU
CAPTION "Synergy"
CLASS "GoSynergy"
FONT 8, "MS Sans Serif"
BEGIN
    LTEXT           "Choose to share or use a shared keyboard and mouse, provide the requested information, then click Test to check your settings or Start to save your settings and start Synergy.",
                    IDC_STATIC,7,7,286,19
    GROUPBOX        "",IDC_STATIC,7,29,286,36
    GROUPBOX        "",IDC_STATIC,7,72,286,36
    GROUPBOX        "Options",IDC_STATIC,7,115,286,56
    CONTROL         "&Use another computer's shared keyboard and mouse (client)",
                    IDC_MAIN_CLIENT_RADIO,"Button",BS_AUTORADIOBUTTON | 
                    WS_GROUP | WS_TABSTOP,11,29,205,10
    CONTROL         "Share this computer's keyboard and mouse (server)",
                    IDC_MAIN_SERVER_RADIO,"Button",BS_AUTORADIOBUTTON,11,72,
                    177,10
    LTEXT           "Other Computer's &Host Name:",
                    IDC_MAIN_CLIENT_SERVER_NAME_LABEL,12,46,94,8
    EDITTEXT        IDC_MAIN_CLIENT_SERVER_NAME_EDIT,111,44,106,12,
                    ES_AUTOHSCROLL
    LTEXT           "&Screens && Links:",IDC_MAIN_SERVER_SCREENS_LABEL,12,89,
                    54,8
    PUSHBUTTON      "Configure...",IDC_MAIN_SCREENS,71,86,50,14
    PUSHBUTTON      "&Options...",IDC_MAIN_OPTIONS,12,129,50,14
    PUSHBUTTON      "Hot &Keys...",IDC_MAIN_HOTKEYS,68,129,50,14
    PUSHBUTTON      "Adva&nced...",IDC_MAIN_ADVANCED,124,129,50,14
    PUSHBUTTON      "&AutoStart...",IDC_MAIN_AUTOSTART,180,129,50,14
    LTEXT           "&Logging Level:",IDC_STATIC,12,154,48,8
    COMBOBOX        IDC_MAIN_DEBUG,68,151,61,60,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "&Info",IDC_MAIN_INFO,7,178,50,14
    PUSHBUTTON   "&Test",IDC_MAIN_TEST,131,179,50,14
    DEFPUSHBUTTON      "Start",IDOK,187,179,50,14
    PUSHBUTTON      "Quit",IDCANCEL,243,179,50,14
END

IDD_ADD DIALOG DISCARDABLE  0, 0, 192, 254
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION
CAPTION "Add Screen"
FONT 8, "MS Sans Serif"
BEGIN
    LTEXT           "&Screen Name:",IDC_STATIC,7,9,46,8
    EDITTEXT        IDC_ADD_SCREEN_NAME_EDIT,79,7,106,12,ES_AUTOHSCROLL
    LTEXT           "&Aliases:",IDC_STATIC,7,25,25,8
    EDITTEXT        IDC_ADD_ALIASES_EDIT,79,26,106,24,ES_MULTILINE | 
                    ES_AUTOVSCROLL | ES_AUTOHSCROLL | ES_WANTRETURN
    GROUPBOX        "Options",IDC_STATIC,7,55,178,54
    LTEXT           "If your Caps, Num, or Scroll Lock keys behave strangely on this client screen then try turning the half-duplex options on and reconnect the client.",
                    IDC_STATIC,13,65,165,25
    CONTROL         "&Caps Lock",IDC_ADD_HD_CAPS_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,13,93,51,10
    CONTROL         "&Num Lock",IDC_ADD_HD_NUM_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,69,93,51,10
    CONTROL         "Sc&roll Lock",IDC_ADD_HD_SCROLL_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,125,93,51,10
    GROUPBOX        "Modifiers",IDC_STATIC,7,113,178,65
    LTEXT           "Shift",IDC_STATIC,13,129,15,8
    COMBOBOX        IDC_ADD_MOD_SHIFT,37,126,48,60,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Ctrl",IDC_STATIC,13,144,11,8
    COMBOBOX        IDC_ADD_MOD_CTRL,37,142,48,60,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Alt",IDC_STATIC,13,160,9,8
    COMBOBOX        IDC_ADD_MOD_ALT,37,158,48,60,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Meta",IDC_STATIC,101,128,17,8
    COMBOBOX        IDC_ADD_MOD_META,125,126,48,60,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "Super",IDC_STATIC,101,144,20,8
    COMBOBOX        IDC_ADD_MOD_SUPER,125,142,48,60,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_TABSTOP
    GROUPBOX        "Dead Corners",IDC_STATIC,7,183,178,43
    LTEXT           "Don't switch in these corners:",IDC_STATIC,14,198,52,18
    CONTROL         "",IDC_STATIC,"Static",SS_BLACKFRAME,68,193,47,28
    CONTROL         "",IDC_ADD_DC_TOP_LEFT,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,76,197,16,8
    CONTROL         "",IDC_ADD_DC_TOP_RIGHT,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,98,197,16,8
    CONTROL         "",IDC_ADD_DC_BOTTOM_LEFT,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,76,210,16,8
    CONTROL         "",IDC_ADD_DC_BOTTOM_RIGHT,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,98,210,16,8
    LTEXT           "Size",IDC_STATIC,120,202,14,8
    EDITTEXT        IDC_ADD_DC_SIZE,139,200,40,12,ES_AUTOHSCROLL | ES_NUMBER
    DEFPUSHBUTTON   "OK",IDOK,79,233,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,135,233,50,14
END

IDD_WAIT DIALOG DISCARDABLE  0, 0, 186, 54
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION
CAPTION "Running Test..."
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "Stop",IDOK,129,33,50,14
    LTEXT           "Running synergy.  Press Stop to end the test.",
                    IDC_STATIC,7,7,172,15
END

IDD_AUTOSTART DIALOG DISCARDABLE  0, 0, 195, 189
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Auto Start"
FONT 8, "MS Sans Serif"
BEGIN
    DEFPUSHBUTTON   "Close",IDCANCEL,138,168,50,14
    LTEXT           "Synergy can be configured to start automatically when you log in.  If you have sufficient access rights, you can instead configure synergy to start automatically when your computer starts.",
                    IDC_STATIC,7,7,181,33
    LTEXT           "You have sufficient access rights to install and uninstall Auto Start for all users or for just yourself.",
                    IDC_AUTOSTART_PERMISSION_MSG,7,69,181,17
    LTEXT           "Synergy is configured to start automatically when the system starts.",
                    IDC_AUTOSTART_INSTALLED_MSG,7,93,181,17
    GROUPBOX        "When &You Log In",IDC_STATIC,7,119,84,40
    PUSHBUTTON      "Install",IDC_AUTOSTART_INSTALL_USER,23,133,50,14
    GROUPBOX        "When &Computer Starts",IDC_STATIC,104,119,84,40
    PUSHBUTTON      "Install",IDC_AUTOSTART_INSTALL_SYSTEM,119,134,50,14
    LTEXT           "Synergy can be configured to start automatically when the computer starts or when you log in but not both.",
                    IDC_STATIC,7,43,181,17
END

IDD_GLOBAL_OPTIONS DIALOG DISCARDABLE  0, 0, 207, 354
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Options"
FONT 8, "MS Sans Serif"
BEGIN
    LTEXT           "It's easy to unintentionally switch screens when the pointer is near a screen's edge.  Synergy can prevent switching until certain conditions are met to reduce unintentional switching.",
                    IDC_STATIC,7,7,191,26
    LTEXT           "Synergy can wait to switch until the cursor has been at a screen's edge for some amount of time.",
                    IDC_STATIC,7,37,193,16
    CONTROL         "Switch after waiting",IDC_GLOBAL_DELAY_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,7,59,77,10
    EDITTEXT        IDC_GLOBAL_DELAY_TIME,112,58,45,12,ES_AUTOHSCROLL | 
                    ES_NUMBER
    LTEXT           "ms",IDC_STATIC,159,60,10,8
    LTEXT           "Synergy can switch only when the cursor hits a screen edge twice within some amount of time.",
                    IDC_STATIC,7,77,193,16
    CONTROL         "Switch on double tap within",IDC_GLOBAL_TWO_TAP_CHECK,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,7,99,103,10
    EDITTEXT        IDC_GLOBAL_TWO_TAP_TIME,112,98,45,12,ES_AUTOHSCROLL | 
                    ES_NUMBER
    LTEXT           "ms",IDC_STATIC,159,100,10,8
    LTEXT           "Synergy can be restricted to switch only if certain modifiers are pressed.",
                    IDC_STATIC,7,119,193,19
    CONTROL         "Check clients every",IDC_GLOBAL_HEARTBEAT_CHECK,"Button",
                    BS_AUTOCHECKBOX | WS_TABSTOP,7,220,78,10
    EDITTEXT        IDC_GLOBAL_HEARTBEAT_TIME,112,219,45,12,ES_AUTOHSCROLL | 
                    ES_NUMBER
    LTEXT           "ms",IDC_STATIC,159,221,10,8
    LTEXT           "Synergy can synchronize screen savers across all screens.",
                    IDC_STATIC,7,243,193,8
    CONTROL         "Synchronize screen savers",IDC_GLOBAL_SCREENSAVER_SYNC,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,7,259,101,10
    LTEXT           "Relative mouse moves on secondary screens.",IDC_STATIC,
                    7,280,193,8
    CONTROL         "Use relative mouse moves",IDC_GLOBAL_RELATIVE_MOVES,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,7,296,99,10
    CONTROL         "Don't take foreground window on Windows servers",
                    IDC_GLOBAL_LEAVE_FOREGROUND,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,7,317,177,10
    DEFPUSHBUTTON   "OK",IDOK,94,333,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,150,333,50,14
    LTEXT           "Synergy can periodically check that clients are still alive and connected.  Use this only if synergy doesn't detect when clients disconnect.",
                    IDC_STATIC,7,189,193,24
    CONTROL         "Only switch when shift pressed.",IDC_GLOBAL_NEEDS_SHIFT,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,7,143,115,10
    CONTROL         "Only switch when control pressed.",
                    IDC_GLOBAL_NEEDS_CONTROL,"Button",BS_AUTOCHECKBOX | 
                    WS_TABSTOP,7,155,123,10
    CONTROL         "Only switch when alt pressed.",IDC_GLOBAL_NEEDS_ALT,
                    "Button",BS_AUTOCHECKBOX | WS_TABSTOP,7,167,109,10
END

IDD_ADVANCED_OPTIONS DIALOG DISCARDABLE  0, 0, 230, 186
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Advanced Options"
FONT 8, "MS Sans Serif"
BEGIN
    LTEXT           "Synergy normally uses this computer's name as its screen name.  Enter another name here if you want to use a different screen name.",
                    IDC_STATIC,7,7,216,19
    LTEXT           "Screen &Name:",IDC_STATIC,7,34,46,8
    EDITTEXT        IDC_ADVANCED_NAME_EDIT,63,32,106,12,ES_AUTOHSCROLL
    LTEXT           "Synergy normally uses a particular network port number.  Enter an alternative port here.  (The server and all clients must use the same port number.)",
                    IDC_STATIC,7,56,216,26
    LTEXT           "&Port:",IDC_STATIC,7,90,16,8
    EDITTEXT        IDC_ADVANCED_PORT_EDIT,63,88,40,12,ES_AUTOHSCROLL | 
                    ES_NUMBER
    LTEXT           "The server normally listens for client connections on all network interfaces.  Enter the address of a particular interface to listen on just that interface.",
                    IDC_STATIC,7,110,216,26
    LTEXT           "&Interface:",IDC_STATIC,7,144,31,8
    EDITTEXT        IDC_ADVANCED_INTERFACE_EDIT,63,142,81,12,ES_AUTOHSCROLL
    PUSHBUTTON      "&Defaults",IDC_ADVANCED_DEFAULTS,7,165,50,14
    DEFPUSHBUTTON   "OK",IDOK,118,165,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,173,165,50,14
END

IDD_SCREENS_LINKS DIALOG DISCARDABLE  0, 0, 354, 213
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Screens & Links"
FONT 8, "MS Sans Serif"
BEGIN
    LTEXT           "&Screens:",IDC_STATIC,7,7,29,8
    LISTBOX         IDC_SCREENS_SCREENS,7,18,100,36,LBS_SORT | 
                    LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "+",IDC_SCREENS_ADD_SCREEN,7,57,17,14
    PUSHBUTTON      "-",IDC_SCREENS_REMOVE_SCREEN,28,57,17,14
    PUSHBUTTON      "Edit",IDC_SCREENS_EDIT_SCREEN,49,57,24,14
    LTEXT           "&Links:",IDC_STATIC,7,83,20,8
    LISTBOX         IDC_SCREENS_LINKS,7,94,339,59,LBS_SORT | 
                    LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    EDITTEXT        IDC_SCREENS_SRC_START,7,156,16,12,ES_AUTOHSCROLL | 
                    ES_NUMBER
    LTEXT           "to",IDC_STATIC,25,158,8,8
    EDITTEXT        IDC_SCREENS_SRC_END,33,156,16,12,ES_AUTOHSCROLL | 
                    ES_NUMBER
    LTEXT           "% of the",IDC_STATIC,52,158,27,8
    COMBOBOX        IDC_SCREENS_SRC_SIDE,80,156,48,69,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "of",IDC_STATIC,129,158,8,8
    COMBOBOX        IDC_SCREENS_SRC_SCREEN,139,156,59,53,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_TABSTOP
    LTEXT           "goes to",IDC_STATIC,200,158,24,8
    EDITTEXT        IDC_SCREENS_DST_START,225,156,16,12,ES_AUTOHSCROLL | 
                    ES_NUMBER
    LTEXT           "to",IDC_STATIC,243,158,8,8
    EDITTEXT        IDC_SCREENS_DST_END,251,156,16,12,ES_AUTOHSCROLL | 
                    ES_NUMBER
    LTEXT           "% of",IDC_STATIC,270,158,15,8
    COMBOBOX        IDC_SCREENS_DST_SCREEN,287,156,59,53,CBS_DROPDOWNLIST | 
                    WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "+",IDC_SCREENS_ADD_LINK,7,172,17,14
    PUSHBUTTON      "-",IDC_SCREENS_REMOVE_LINK,28,172,17,14
    DEFPUSHBUTTON   "OK",IDOK,241,192,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,297,192,50,14
    LTEXT           "Source edge overlaps an existing edge.",
                    IDC_SCREENS_OVERLAP_ERROR,72,175,126,8,NOT WS_VISIBLE | 
                    NOT WS_GROUP
END

IDD_INFO DIALOG DISCARDABLE  0, 0, 186, 95
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Info"
FONT 8, "MS Sans Serif"
BEGIN
    LTEXT           "Version:",IDC_STATIC,7,7,26,8
    EDITTEXT        IDC_INFO_VERSION,52,7,127,12,ES_AUTOHSCROLL | 
                    ES_READONLY | NOT WS_BORDER
    LTEXT           "Hostname:",IDC_STATIC,7,19,35,8
    EDITTEXT        IDC_INFO_HOSTNAME,52,19,127,12,ES_AUTOHSCROLL | 
                    ES_READONLY | NOT WS_BORDER
    LTEXT           "IP Address:",IDC_STATIC,7,31,37,8
    EDITTEXT        IDC_INFO_IP_ADDRESS,52,31,127,12,ES_AUTOHSCROLL | 
                    ES_READONLY | NOT WS_BORDER
    LTEXT           "User Config:",IDC_STATIC,7,43,40,8
    EDITTEXT        IDC_INFO_USER_CONFIG,52,43,127,12,ES_AUTOHSCROLL | 
                    ES_READONLY | NOT WS_BORDER
    LTEXT           "Sys Config:",IDC_STATIC,7,55,36,8
    EDITTEXT        IDC_INFO_SYS_CONFIG,52,55,127,12,ES_AUTOHSCROLL | 
                    ES_READONLY | NOT WS_BORDER
    DEFPUSHBUTTON   "OK",IDOK,129,74,50,14
END

IDD_HOTKEY_OPTIONS DIALOG DISCARDABLE  0, 0, 360, 151
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Hot Keys"
FONT 8, "MS Sans Serif"
BEGIN
    LTEXT           "&Hot Keys:",IDC_STATIC,7,7,32,8
    LISTBOX         IDC_HOTKEY_HOTKEYS,7,18,169,88,LBS_SORT | 
                    LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "+",IDC_HOTKEY_ADD_HOTKEY,7,109,17,14
    PUSHBUTTON      "-",IDC_HOTKEY_REMOVE_HOTKEY,28,109,17,14
    PUSHBUTTON      "Edit",IDC_HOTKEY_EDIT_HOTKEY,49,109,24,14
    LTEXT           "&Actions:",IDC_STATIC,183,7,26,8
    LISTBOX         IDC_HOTKEY_ACTIONS,183,18,169,88,LBS_SORT | 
                    LBS_NOINTEGRALHEIGHT | WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "+",IDC_HOTKEY_ADD_ACTION,183,109,17,14
    PUSHBUTTON      "-",IDC_HOTKEY_REMOVE_ACTION,204,109,17,14
    PUSHBUTTON      "Edit",IDC_HOTKEY_EDIT_ACTION,225,109,24,14
    DEFPUSHBUTTON   "OK",IDOK,302,130,50,14
END

IDD_HOTKEY_CONDITION DIALOG DISCARDABLE  0, 0, 183, 58
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Hot Key"
FONT 8, "MS Sans Serif"
BEGIN
    LTEXT           "Enter &new hot key or mouse button:",IDC_STATIC,7,7,113,
                    8
    EDITTEXT        IDC_HOTKEY_CONDITION_HOTKEY,7,17,169,12,ES_WANTRETURN
    DEFPUSHBUTTON   "OK",IDOK,70,37,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,126,37,50,14
END

IDD_HOTKEY_ACTION DIALOG DISCARDABLE  0, 0, 183, 218
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Action"
FONT 8, "MS Sans Serif"
BEGIN
    LTEXT           "&Action:",IDC_STATIC,7,7,23,8
    CONTROL         "Press:",IDC_HOTKEY_ACTION_DOWN,"Button",
                    BS_AUTORADIOBUTTON | WS_TABSTOP,7,19,35,10
    CONTROL         "Release:",IDC_HOTKEY_ACTION_UP,"Button",
                    BS_AUTORADIOBUTTON,7,31,44,10
    CONTROL         "Press && Release:",IDC_HOTKEY_ACTION_DOWNUP,"Button",
                    BS_AUTORADIOBUTTON,7,43,69,10
    CONTROL         "Switch To Screen:",IDC_HOTKEY_ACTION_SWITCH_TO,"Button",
                    BS_AUTORADIOBUTTON,7,85,75,10
    CONTROL         "Switch In Direction:",IDC_HOTKEY_ACTION_SWITCH_IN,
                    "Button",BS_AUTORADIOBUTTON,7,101,77,10
    CONTROL         "Lock Cursor to Screen:",IDC_HOTKEY_ACTION_LOCK,"Button",
                    BS_AUTORADIOBUTTON,7,117,89,10
    CONTROL         "Keyboard broadcasting:",
                    IDC_HOTKEY_ACTION_KEYBOARD_BROADCAST,"Button",
                    BS_AUTORADIOBUTTON,7,133,89,10
    LTEXT           "&Hot key or mouse button:",IDC_STATIC,7,55,80,8
    EDITTEXT        IDC_HOTKEY_ACTION_HOTKEY,7,67,152,12,ES_WANTRETURN
    PUSHBUTTON      "...",IDC_HOTKEY_ACTION_SCREENS,162,67,14,12
    COMBOBOX        IDC_HOTKEY_ACTION_SWITCH_TO_LIST,87,83,89,62,
                    CBS_DROPDOWNLIST | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_HOTKEY_ACTION_SWITCH_IN_LIST,106,99,70,66,
                    CBS_DROPDOWNLIST | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_HOTKEY_ACTION_LOCK_LIST,106,115,70,58,
                    CBS_DROPDOWNLIST | WS_VSCROLL | WS_TABSTOP
    COMBOBOX        IDC_HOTKEY_ACTION_KEYBOARD_BROADCAST_LIST,106,131,53,58,
                    CBS_DROPDOWNLIST | WS_VSCROLL | WS_TABSTOP
    PUSHBUTTON      "...",IDC_HOTKEY_ACTION_KEYBOARD_BROADCAST_SCREENS,
                    162,131,14,12
    LTEXT           "Action takes place &when:",IDC_STATIC,7,153,81,8
    CONTROL         "Hot key is pressed",IDC_HOTKEY_ACTION_ON_ACTIVATE,
                    "Button",BS_AUTORADIOBUTTON | WS_TABSTOP,7,165,74,10
    CONTROL         "Hot key is released",IDC_HOTKEY_ACTION_ON_DEACTIVATE,
                    "Button",BS_AUTORADIOBUTTON,7,177,76,10
    DEFPUSHBUTTON   "OK",IDOK,70,197,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,126,197,50,14
END

IDD_HOTKEY_SCREENS DIALOG DISCARDABLE  0, 0, 237, 79
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Target Screens"
FONT 8, "MS Sans Serif"
BEGIN
    LTEXT           "&Available screens:",IDC_STATIC,7,7,58,8
    LISTBOX         IDC_HOTKEY_SCREENS_SRC,7,17,100,36,LBS_NOINTEGRALHEIGHT | 
                    LBS_EXTENDEDSEL | WS_VSCROLL | WS_TABSTOP
    LISTBOX         IDC_HOTKEY_SCREENS_DST,130,17,100,36,
                    LBS_NOINTEGRALHEIGHT | LBS_EXTENDEDSEL | WS_VSCROLL | 
                    WS_TABSTOP
    PUSHBUTTON      "-->",IDC_HOTKEY_SCREENS_ADD,109,21,17,14
    PUSHBUTTON      "<--",IDC_HOTKEY_SCREENS_REMOVE,109,38,17,14
    DEFPUSHBUTTON   "OK",IDOK,124,58,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,180,58,50,14
    LTEXT           "&Send action to screens:",IDC_STATIC,130,7,76,8
END


/////////////////////////////////////////////////////////////////////////////
//
// DESIGNINFO
//

#ifdef APSTUDIO_INVOKED
GUIDELINES DESIGNINFO DISCARDABLE 
BEGIN
    IDD_MAIN, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 293
        TOPMARGIN, 7
        BOTTOMMARGIN, 192
    END

    IDD_ADD, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 185
        TOPMARGIN, 7
        BOTTOMMARGIN, 247
    END

    IDD_WAIT, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 179
        TOPMARGIN, 7
        BOTTOMMARGIN, 47
    END

    IDD_AUTOSTART, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 188
        TOPMARGIN, 7
        BOTTOMMARGIN, 182
    END

    IDD_GLOBAL_OPTIONS, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 200
        TOPMARGIN, 7
        BOTTOMMARGIN, 347
    END

    IDD_ADVANCED_OPTIONS, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 223
        TOPMARGIN, 7
        BOTTOMMARGIN, 179
    END

    IDD_SCREENS_LINKS, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 347
        TOPMARGIN, 7
        BOTTOMMARGIN, 206
    END

    IDD_INFO, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 179
        TOPMARGIN, 7
        BOTTOMMARGIN, 88
    END

    IDD_HOTKEY_OPTIONS, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 353
        TOPMARGIN, 7
        BOTTOMMARGIN, 144
    END

    IDD_HOTKEY_CONDITION, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 176
        TOPMARGIN, 7
        BOTTOMMARGIN, 51
    END

    IDD_HOTKEY_ACTION, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 176
        TOPMARGIN, 7
        BOTTOMMARGIN, 195
    END

    IDD_HOTKEY_SCREENS, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 230
        TOPMARGIN, 7
        BOTTOMMARGIN, 72
    END
END
#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
IDI_SYNERGY             ICON    DISCARDABLE     "synergy.ico"

/////////////////////////////////////////////////////////////////////////////
//
// String Table
//

STRINGTABLE DISCARDABLE 
BEGIN
    IDS_ERROR               "Error"
    IDS_INVALID_SCREEN_NAME "Screen name `%{1}' is invalid."
    IDS_DUPLICATE_SCREEN_NAME "The screen name `%{1}' is already being used."
    IDS_SCREEN_NAME_IS_ALIAS "A name may not be an alias of itself."
    IDS_VERIFY              "Confirm"
    IDS_UNSAVED_DATA_REALLY_QUIT "You have unsaved changes.  Really quit?"
    IDS_UNKNOWN_SCREEN_NAME "The screen name `%{1}' is not in the layout."
    IDS_INVALID_PORT        "The port `%{1}' is invalid.  It must be between 1 and 65535 inclusive.  %{2} is the standard port."
    IDS_SAVE_FAILED         "Failed to save configuration: %{1}"
    IDS_STARTUP_FAILED      "Failed to start synergy: %{1}"
    IDS_STARTED_TITLE       "Started"
    IDS_STARTED             "Synergy was successfully started.  Use the task manager or tray icon to terminate it."
    IDS_UNINSTALL_TITLE     "Removed Auto-Start"
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDS_AUTOSTART_PERMISSION_SYSTEM 
                            "You have sufficient access rights to install and uninstall Auto Start for all users."
    IDS_AUTOSTART_PERMISSION_USER 
                            "You have sufficient access rights to install and uninstall Auto Start for just yourself."
    IDS_AUTOSTART_PERMISSION_NONE 
                            "You do not have sufficient access rights to install or uninstall Auto Start."
    IDS_AUTOSTART_INSTALLED_SYSTEM 
                            "Synergy is configured to start automatically when the system starts."
    IDS_AUTOSTART_INSTALLED_USER 
                            "Synergy is configured to start automatically when you log in."
    IDS_AUTOSTART_INSTALLED_NONE 
                            "Synergy is not configured to start automatically."
    IDS_INSTALL_LABEL       "Install"
    IDS_UNINSTALL_LABEL     "Uninstall"
    IDS_INSTALL_GENERIC_ERROR "Install failed: %{1}"
    IDS_UNINSTALL_GENERIC_ERROR "Uninstall failed: %{1}"
    IDS_INSTALL_TITLE       "Installed Auto-Start"
    IDS_INSTALLED_SYSTEM    "Installed auto-start.  Synergy will automatically start each time you start your computer."
    IDS_INSTALLED_USER      "Installed auto-start.  Synergy will automatically start each time you log in."
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDS_UNINSTALLED_SYSTEM  "Removed auto-start.  Synergy will not automatically start each time you start or reboot your computer."
    IDS_UNINSTALLED_USER    "Removed auto-start.  Synergy will not automatically start each time you log in."
    IDS_INVALID_SERVER_NAME "Server name `%{1}' is invalid."
	IDS_TITLE               "%{1} Version %{2}"
    IDS_SERVER_IS_CLIENT    "Please enter the name of the computer sharing a\nkeyboard and mouse, not the name of this computer,\nin the Other Computer's Host Name field."
    IDS_ADD_SCREEN          "Add Screen"
    IDS_EDIT_SCREEN         "Edit Screen %{1}"
    IDS_ERROR_CODE          "Error code: %{1}"
    IDS_AUTOSTART_PERMISSION_ALL 
                            "You have sufficient access rights to install and uninstall Auto Start for all users or for just yourself."
    IDS_INVALID_INTERFACE_NAME "The interface '%{1}' is invalid: %{2}"
    IDS_INVALID_CORNER_SIZE "The dead corner size %{1} is invalid;  it must be 0 or higher."
    IDS_NEW_LINK            "[New Link]"
    IDS_SIDE_LEFT           "left of"
    IDS_SIDE_RIGHT          "right of"
    IDS_SIDE_TOP            "above"
    IDS_SIDE_BOTTOM         "below"
END

STRINGTABLE DISCARDABLE 
BEGIN
    IDS_LINK_FORMAT         "%{4}%{5} is %{3} %{1}%{2}"
    IDS_LINK_INTERVAL_FORMAT "(%{1},%{2})"
    IDS_EDGE_LEFT           "left"
    IDS_EDGE_RIGHT          "right"
    IDS_EDGE_TOP            "top"
    IDS_EDGE_BOTTOM         "bottom"
    IDS_AUTOSTART_SAVE_FAILED "Failed to save autostart configuration: %{1}"
    IDS_LOAD_FAILED         "Failed to load configuration."
    IDS_CONFIG_CHANGED      "Configuration changed on disk.  Reload?"
    IDS_MODE_OFF            "off"
    IDS_MODE_ON             "on"
    IDS_MODE_TOGGLE         "toggle"
    IDS_ALL_SCREENS         "All Screens"
    IDS_ACTIVE_SCREEN       "Active Screen"
END

#endif    // English (U.S.) resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//


/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

