/*
 * synergy -- mouse and keyboard sharing utility
 * Copyright (C) 2011 Chris Schoeneman, Nick Bolton, Sorin Sbarnea
 * 
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * found in the file COPYING that should have accompanied this file.
 * 
 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef ELEVEL_H
#define ELEVEL_H

//! Log levels
/*!
The logging priority levels in order of highest to lowest priority.
*/
enum ELevel {
	kPRINT = -1,	//!< For print only (no file or time)
	kFATAL,			//!< For fatal errors
	kERROR,			//!< For serious errors
	kWARNING,		//!< For minor errors and warnings
	kNOTE,			//!< For messages about notable events
	kINFO,			//!< For informational messages
	kDEBUG,			//!< For important debugging messages
	kDEBUG1,		//!< For more detailed debugging messages
	kDEBUG2			//!< For even more detailed debugging messages
};

#endif
